package cs297.xmldb.src.bm;
import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.server.*;

/**
 Manage pool of pages. Called BM at initialize, the number 
 of pages in the pool is specified in properties file.
 As pages are used up, this is when free list is empty, 
 this class will flush the oldest page on to disk and 
 allocate new, empty page into buffer pool.   
 @Date: Spring & fall 2004
 @Author: Thien An Nguyen    
 */
public class ThBufferMgr
{
   public static int pSn;  /* Global varible for page number */
   public Hashtable bufferPool = new Hashtable(); /* buffer pool */
   public ArrayList freeList = new ArrayList(); /* number of the free pages */
   public ArrayList leastUse = new ArrayList(); /* number of oldest pages */
  
   public ArrayList freeStrucList = new ArrayList();  /* number of the free page*/
   public ArrayList leastStrucUse = new ArrayList();  /*number of oldest page*/
  
   public int poolSize;     /* total page in the pool*/
   public int strucPageNum; /* number of space map page in the pool*/
   public int pSize;

  /**
   Buffer Manager will be construct at db sub-system starts up
    1. At start up, BM construct buffer pool of x pages 
   (specified in dbsystem.properties)
    2. These pages are also added in the free list
   @param Properties values
   @return:  none (A Buffer Manager is created).
   */ 
  public ThBufferMgr(Properties prop)
  {
     poolSize = (new Integer(prop.getProperty("poolSize"))).intValue();
     pSize = (new Integer(prop.getProperty("pageSize"))).intValue();
     pSn = (new Integer(prop.getProperty("startPageNum"))).intValue();
     strucPageNum = (new Integer(prop.getProperty("numberOfStrucPage"))).intValue();
     int endNum = pSn + strucPageNum; //number of space map page
     //add the number of space map page in buffer pool
     while(pSn<endNum)
     {
        bufferPool.put(Integer.toString(pSn), new ThStructMapPage(pSize, pSn));
        freeStrucList.add(Integer.toString(pSn));
        pSn++;
     }             
     
     /* total page in buffer pool */
     endNum = pSn + poolSize - strucPageNum;  
     
     /* Now, fill the remaining space with data pages.
     There are 2 kind of pages because we are experimenting with 2 
     kind of recovery methods: ThPage and ThPageInterpreter. */
    
     /* Check to see which recovery method specified */
     String rmethod = (String)prop.get("recoverMethode");
     if (rmethod.equals("NATIX"))
     { //then create free list of logical Pagerinterpreters in buffer pool
        while(pSn<endNum)
        {
           bufferPool.put(Integer.toString(pSn), 
                           new ThPageInterpreter(pSize, pSn));
           freeList.add(Integer.toString(pSn));
           pSn++;
        }              
     } 
     else 
     { //then create free list of logical pages in buffer pool
        while(pSn<endNum)
        {
           bufferPool.put(Integer.toString(pSn), new ThPage(pSize, pSn));
           freeList.add(Integer.toString(pSn));
           pSn++;
        }      
     } 
  }  
  
  /**
   This function call SM to load a page from stable storage.
   @param Request page number, 
   @param ThSystem so that BM can call SM to retrieve 
   the page on disk
   @return Requested page
   */
  public IThPage loadPage(String pagenum, ThSystem sys)
  {  
     return sys.getSM().loadPage(pagenum, pSize, sys); 
  }  
  
  /**
   This function call SM to flush a page from stable storage.
   @param Request page number, 
   @param ThSystem so that BM can call SM to retrieve the page on disk
   @return:  physical location (physical ID) of the page   
   */
  public String flush(IThPage page, ThSystem sys)
  {  
    if (sys.isNatix() &&
        page.getClass().getName().indexOf("ThPageInterpreter") >= 0)
       ((ThPageInterpreter)page).publicLogsToLM(sys);
    
    return sys.getSM().flushPage(page);
  }
  
  /**
   This function call SM to flush all page in buffer pool to disk.
   @param ThSystem so that BM can call SM to retrieve the page on disk.
   @return:  0 is success.
   */
  public int flushAllPage(ThSystem sys)
  {  
     ThStorageMgr sm = sys.getSM();
     Enumeration keys = bufferPool.keys();
    
     while (keys.hasMoreElements())
     {
        String key = (String)keys.nextElement();
        IThPage page = (IThPage)bufferPool.get(key);
        flush(page, sys);
     }
    
     return 0; // success
  }
  
  /**
   This function will be called only by ThPageInterpreter
   to send it's local log records to Log Manager (NATIX's feature.)
   @param  ThSystem where all the global variables hold.
   @Return:  Return Code.
   */
  public int publicLogs(ThSystem sys)
  {
     ThStorageMgr sm = sys.getSM();
     Enumeration keys = bufferPool.keys();
    
     while (keys.hasMoreElements())
     {
        String key = (String)keys.nextElement();
        IThPage page = (IThPage)bufferPool.get(key);
        if (sys.isNatix() &&
           page.getClass().getName().indexOf("ThPageInterpreter") >= 0)
         ((ThPageInterpreter)page).publicLogsToLM(sys);
     }
         
     return 0;
  }
  
  /**
   If the page is in the buffer: then return the page
   else if the page is not in the buffer write the least use page 
   to disk. And load the page from disk.
   @Receive: The requested page number
   @param ThSystem where global variables hold.
   @Return:  The requested page.
   */
  public IThPage getPage(String key, ThSystem sys) 
  {    
     if (bufferPool.containsKey(key))
        return (IThPage)bufferPool.get(key);
    
     if (freeList.size() < 1)
     { //write the least use page (block) to disk. 
        String firstKey = (String) leastUse.remove(0);
        //remove the page from buffer
        IThPage page = (IThPage)bufferPool.remove(firstKey);
        flush(page, sys);
        page = null;
        //load page from disk.
        page = loadPage(key, sys);
        if (page != null)
        {
           bufferPool.put(key, page);
           leastUse.add(key);
           return page;
        }
    }
    else
    {      
       freeList.remove(0);
       IThPage page = loadPage(key, sys);
       bufferPool.put(key, page);
       leastUse.add(key);
       return page;
    }
    
    return null;
  }
  
  /**
   Get last updated page in buffer.  
   If the page has enough free space as the specifed length, return the page.
   else if the page doesn't have enough free space, flush the page onto disk, 
   and allocate new page
   @param The length of free space.
   @return:  The allocated page.
   */
  public IThPage getPage(int datlength, ThSystem sys)
  {
     String key;
     IThPage page;
    
     if (leastUse.size() < 1)
     {
        key = (String)freeList.remove(0);
        leastUse.add(key);
        return (IThPage)bufferPool.get(key);
     }
    
     key = (String)leastUse.get(leastUse.size()-1);
     page = (IThPage)bufferPool.get(key);
    
     if (page.isFull(datlength))
        page = allocatePage(sys);
    
     return page;
  }
  
  /**
   If there is free page is in the buffer: then return the page
   else if the page is not in the buffer write the oldest page to
   disk and allocate new page
   @param ThSystem where global variables held
   @return:  A new, empty page.
   */
  public IThPage allocatePage(ThSystem sys)
  {
     if (freeList.size() < 1)
     {
        String firstKey = (String) leastUse.remove(0);
        //remove the page from buffer
        IThPage page = (IThPage)bufferPool.remove(firstKey);
        flush(page, sys); //will come back. right now null
        page = null;
      
        if (sys.isNatix())
           page = new ThPageInterpreter(pSize, pSn);
        else
           page = new ThPage(pSize, pSn);
        bufferPool.put(Integer.toString(pSn), page);
        leastUse.add(Integer.toString(pSn));
        pSn++;
      
        return page;
     } 
    
     /* remove the 1st key in freelist */
     String key = (String) freeList.remove(0);  
     leastUse.add(key);
     
     return (IThPage)bufferPool.get(key);
  }
  
  /**
   If there is free page is in the buffer: then return the page
   else if the page is not in the buffer write the oldest page to 
   disk and allocate new page
   @param ThSystem where global variables held
   @return:  A new, empty page.
   */
  public IThPage allocateStrucPage(ThSystem sys)
  {
     if (freeStrucList.size() < 1)
     {
         String firstKey = (String) leastStrucUse.remove(0);
         //remove the page from buffer
         IThPage page = (IThPage)bufferPool.remove(firstKey);
         flush(page, sys); 
         page = new ThStructMapPage(pSize, pSn);
      
        bufferPool.put(Integer.toString(pSn), page);
        leastStrucUse.add(Integer.toString(pSn));
        pSn++;
      
        return page;
     } 
    
     //remove the 1st key in freelist
     String key = (String) freeStrucList.remove(0);  
     leastStrucUse.add(key);
     return (IThPage)bufferPool.get(key);
  }  
  
  /**
   If there is free page is in the buffer: then return the page
   else if the page is not in the buffer write the oldest page to
   disk and allocate new page
   @param ThSystem where global variables held
   @return:  A new, empty page.
   */
  public int allocateStrucPageNum(ThSystem sys)
  {
     if (freeStrucList.size() < 1)
     {
        String firstKey = (String) leastStrucUse.remove(0);
        //remove the page from buffer
        IThPage page = (IThPage)bufferPool.remove(firstKey);
        flush(page, sys); 
        page = new ThStructMapPage(pSize, pSn);
      
        bufferPool.put(Integer.toString(pSn), page);
        leastStrucUse.add(Integer.toString(pSn));
        pSn++;
      
        return pSn-1;
     } 
    
     //remove the 1st key in freelist
     String key = (String) freeStrucList.remove(0);  
     leastStrucUse.add(key);
     return Integer.parseInt(key);
  }  
  
  /**
   This function is called to free all pages in the buffer pool
   when the pages are no longer contains valid data. This 
   function only called  after a LOAD operation.
   @param ThSystem where global vairables held.
   @return:  Return Code for this operation.
   */
  public int freeAllPages(ThSystem sys)
  {
     ThStorageMgr sm = sys.getSM();
     Enumeration keys = bufferPool.keys();
     Properties prop = sys.getProperties();
    
     while (keys.hasMoreElements())
     {
        String key = (String)keys.nextElement();
        IThPage page = (IThPage)bufferPool.remove(key);
        sm.freePage(page);
     }
     sm.cleanBMDirectory();
     freeList = new ArrayList();  //the number of the free page
     leastUse = new ArrayList(); 
     //load new page in bufferpool
     pSn = (new Integer(prop.getProperty("startPageNum"))).intValue();
     int endNum = pSn + strucPageNum; //number of space map page
     //add the number of space map page in buffer pool
     while(pSn<endNum)
     {
        bufferPool.put(Integer.toString(pSn), new ThStructMapPage(pSize, pSn));
        freeStrucList.add(Integer.toString(pSn));
        pSn++;
     }             
    
     endNum = pSn + poolSize - strucPageNum;  //total page in buffer pool
     /*Now, fill the remaining space with data pages.
     There are 2 kind of pages because we are experimenting with 2 
     kind of recovery methods: ThPage and ThPageInterpreter. */
    
     //Check to see which recovery method specified
     String rmethod = (String)prop.get("recoverMethode");
     if (rmethod.equals("NATIX"))
     { //then create free list of logical Pagerinterpreters in buffer pool
        while(pSn<endNum)
        {
           bufferPool.put(Integer.toString(pSn), new ThPageInterpreter(pSize, pSn));
           freeList.add(Integer.toString(pSn));
           pSn++;
        }              
     } 
     else 
     { //then create free list of logical pages in buffer pool
        while(pSn<endNum)
        {
           bufferPool.put(Integer.toString(pSn), new ThPage(pSize, pSn));
           freeList.add(Integer.toString(pSn));
           pSn++;
        }      
     } 
    
     return 0; 
  }
}