package cs297.xmldb.src.bm;
import java.io.*;
import java.util.*;
import java.lang.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.server.*; 
/**
 * Date: Spring & fall 2004
 * @author: Thien An Nguyen, 
 * Purpose: A data page is a storage unit for a database 
 *           to stores it's data. 
 *          When the page is in memory, it is exist in
 *           a high level structured form.
 *          When the page need to be flush onto disk,
 *           it is write out as a block or chunk
 *           (K size) of hex data.     
 *
 * A page contains:
 *  1. page header
 *  2. data records
 *  3. record node headers
 *
 * Page header contains: totals of 17 bytes
 *  1. 2 bytes: page number
 *  2. 1 byte: information such as dirty bit
 *  3. 2 bytes: page LSN (lsn of the last update to this page)
 *
 * For each header record:
 *  2 bytes record id
 *  1 byte record info. such as rec. type, child or not
 *  2 bytes offset to data record in this page
 *
 * For each data record:
 * 2 bytes record length
 * The acctual data
 */
public class ThPage implements IThPage
{
  int pSize; //page size
  int headerSize = 20; //page header size
  int pNum;     //2 bytes page num
  BitSet inFo;  //1 byte (bit 7 - on if this page is dirty)
  int pageLSN;  //2 bytes LSN of the last updated tx
  int freeSpace; //free space
  ArrayList nodeS; //pointers to nodes in page 
  
   /**
   This constructor construct a page in memmory 
   @param the size for the page (specified in properties file)
   @Return: None (A page is created)
   */
   public ThPage(int size, int pnum)
   {
      pSize = size;
      pNum = pnum; 
      pageLSN = 0; //lsn of the last update
      inFo = new BitSet(8); //not dirty at create
      //size - 4 bytes pnum - 1 byte dirty - 4 bytes LSN - 4 currIdPos - 4 deleted
      freeSpace = size - headerSize; //number of bytes free
      nodeS = new ArrayList();
   }
  
   /**
   This constructor create a page in memmory by loading it's 
   information from disk
   1. The size for this page (specified in properties file)
   2. Data block/chunk stored on disk
   @return: None (A page is created)
   */
   public ThPage(int size, byte[] bytes)
   {
      final int byte4 = 4;
      int pos = 0;  
      int backpos = size;
      freeSpace = size - headerSize;
    
      pSize = size;
    
     //1. Header page
     pNum = ThUtilities.hexToNumber(bytes, pos); //get Page number
     pos += byte4;
     inFo = ThUtilities.CharToBitset(bytes, pos);
     pos++;

     nodeS = new ArrayList();
     int id;
     int offset;
     int len;
     ArrayList attr;
     String tag, data;
     BitSet info;
     ThNode node;
    
     int num = ThUtilities.hexToNumber(bytes, pos); //# of nodes
     for (int i=0; i<num; i++)
     {
        backpos -= byte4; // id(4)
        id = ThUtilities.hexToNumber(bytes, backpos);
        backpos -= 1; //bitset
        info = ThUtilities.CharToBitset(bytes, backpos);
      
        backpos -= byte4; // offset(4)
        offset = ThUtilities.hexToNumber(bytes, backpos);
        freeSpace -= 9;
      
        attr = new ArrayList();
        if (info.get(7)) //g node
        {
           len = ThUtilities.hexToNumber(bytes, offset);
           offset += byte4; //len of tag
           tag = ThUtilities.hexToData(bytes, offset, len);
           offset += len;
           attr.add(tag);
           freeSpace -= (4 + len);
        
           len = ThUtilities.hexToNumber(bytes, offset);
           offset += byte4; //len of data
           data = ThUtilities.hexToData(bytes, offset, len);
           offset += len; // don't need this stmt
           attr.add(data);
        
           node = new ThNode(id, true, attr);
           freeSpace -= (4 + len);
       }
       else 
       {
          len = ThUtilities.hexToNumber(bytes, offset);
          offset += byte4; //len of data
          data = ThUtilities.hexToData(bytes, offset, len);
          offset += len;
          attr.add(data);
        
          node = new ThNode(id, false, attr);      
          freeSpace -= (4 + len);
       }
       
       nodeS.add(node);
     }
  }
  
   /**
   This function is called when this page is need to be flushed on 
   a disk.
   Data on this page will be converted into hex characters before 
   this function returns.
   For a number, i.e. pagenum, offset,...,
   it's converted to hex using 2 bytes.
   The convertion is done by ThUtilities class.
   For a character, i.e. character data, 
   it's converted to hex using 1 byte.
   The convertion is done by ThUtilities class.
   @param nothing
   @return character array (in hex) of specified page size.
   */
   public char[] toHex()
   {
      int backpos = pSize;
      int pos = 0;
      char[] arr = new char[pSize];
    
      /* 1. page header */
      /* store 2 bytes page number, pos = 0 */
      pos += ThUtilities.numberToHex(arr, pNum, pos); 
    
      /* store 1 byte dirty bit, pos = 4 */
      pos += ThUtilities.BitsetToChar(arr, inFo, pos);
   
      /* store 2 bytes number of nodes */
      int nnodes = nodeS.size();  
      pos += ThUtilities.numberToHex(arr, nnodes, pos);     

     /* 2. id blocks */
     ThNode node;
     ArrayList attr;
     int id;
     String data;
     BitSet info;

     for (int i=0;i<nnodes;i++) /* write all the node headers */
     {
        node = (ThNode) nodeS.get(i);
        /* store node id */
        id = node.getId();
        attr = node.getData();
        info = node.getInfoByte();
      
        backpos -= 4;
        ThUtilities.numberToHex(arr, id, backpos); // 2 bytes for id
        backpos -= 1;
        ThUtilities.BitsetToChar(arr, info, backpos); 
        backpos -= 4;
        /* store offset to data 2 bytes */
        ThUtilities.numberToHex(arr, pos, backpos);

        /* store node data */
        data = (String) attr.get(0);
        pos += ThUtilities.numberToHex(arr, data.length()*2, pos);
        pos += ThUtilities.dataToHex(arr, data, pos);
        if (info.get(7))
        {
           data = (String) attr.get(1);
           pos += ThUtilities.numberToHex(arr, data.length()*2, pos);
           pos += ThUtilities.dataToHex(arr, data, pos);
       }
     }
    
     return arr;
   }  
  
   public int getPageNum() 
   {
      return pNum; 
   }
  
  /**
   1. This function will be called by insert method
      of a page to format for an insert log record.
   2. Calls LM to write formatted log record.
   3. Calls TM to write the inserted lsn (number).
   @param log record type (String)
   @param transaction id (String)
   @param id of the record (int)
   @param path of the inserting data (String)
   @param data record(s) to be inserted (ArrayList)
   @param name of the xml document (String)
   @param name of the database (String)
   @param ThSystem where all the global variables hold.
   @return: lsn of the inserted record
   */
   public String createInsertLogRecord(String rectype, String txid,
            int id, String path, ArrayList values, String xmlname, 
            String dbname, ThSystem sys)
   {
      String lsn = sys.getLSN();
    
      String data = rectype        + " "     + lsn + " "
                  + dbname + " " + xmlname + " "
                  + getPageNum() + " " 
                  + txid   + " " + Integer.toString(id)+" "
                  + path   + " " + "(";    
      data += (String) values.get(0);    
      for (int i=1; i<values.size(); i++)
         data += " "+(String) values.get(i);   
      data += ")";    

      sys.getLM().writeLogRecord(data);    
      sys.getTM().addUndoLSN(txid, lsn);
    
      return lsn;
   }
  
   /**
   1. Format a log record for a delete statement
   2. Calls LM to write formatted log record.
   3. Calls TM to write the inserted lsn (number).
   @param log record type (String)
   @param transaction id (String)
   @param id of the record (int)
   @param path of the deleting data (String)
   @param name of the xml document (String)
   @param name of the database (String)
   @param system where all the global variables hold.
   @return: 
   */
   public String createDelLogRecord(String rectype, String txid, 
         int id, String path, String xmlname, String dbname, ThSystem sys)
   {
      String lsn = sys.getLSN();
    
      String data = rectype + " "+ lsn + " "
                + dbname + " " + xmlname + " "
                + getPageNum() + " " 
                + txid   + " " + Integer.toString(id)+" "
                + path;

      sys.getLM().writeLogRecord(data);    
      sys.getTM().addUndoLSN(txid, lsn); 
    
      return lsn;
   }  
  
   /**
   1. Format for an update log record
   2. Calls LM to write formatted log record.
   3. Calls TM to write the inserted lsn (number).
   @param log record type (String)
   @param transaction id (String)
   @param id of the record (int)
   @param path of the inserting data (String)
   @param data value before update
   @param data value after update
   @param name of the xml document (String)
   @param name of the database (String)
   @param system where all the global variables hold.
   @return: lsn of the inserted record
   */
   public String createUpdateLogRecord(String rectype, String txid,
             int id, String path, String oldvalue, String newvalue,
             String xmlname, String dbname, ThSystem sys)
   {
      String lsn = sys.getLSN();
    
      String data = rectype        + " "     + lsn + " "
                  + dbname + " " + xmlname + " "
                  + getPageNum() + " " 
                  + txid   + " " + Integer.toString(id)+" "
                  + " "    + path
                  + " ("    +oldvalue+")"
                  + " ("    +newvalue+")";

      sys.getLM().writeLogRecord(data);    
      sys.getTM().addUndoLSN(txid, lsn);
    
      return lsn;
   } 
  
   /**
   Insert a record into this data page.
   @param Type of insert, i.e. insert leave or group.
   @param Transaction id
   @param Id of the record
   @param Inserting path
   @param List of data records to be inserted
   @param Name of the inserting xml document
   @param Name of the database
   @param True if inserting record is a group, False otherwise.
   @param System where all the global variables hold.
   @param True if this function was called by LOAD operation,
          False if this function was called by regular INSERT.
   @return 0 if operation success will through exception is not success 
   */
   public int insert(String msg, String txid, int id, String path, 
                   ArrayList value, String xmlname, String dbname, 
                   boolean isgroup, ThSystem sys, boolean natixlog, 
                   boolean islog)
   {
      inFo.set(7); // set dirty bit on
      String lsn = "";
      if (islog)
      {
         lsn = createInsertLogRecord(msg, txid, id, path, 
                    value, xmlname, dbname, sys);
         pageLSN = Integer.parseInt(lsn); // update pageLSN
      }
       
      ThNode node = new ThNode(id, isgroup, value);
      nodeS.add(node);
      
      return 0; //success
   }  

   public int insert(IThNode node, int parentid, ThSystem sys) 
   {
      return -1;
   }
    
   public int insert(IThNode newstructednode, IThNode prev, ThSystem sys)
   {
      return -1;
   }
  
   public static int computeGRecLen(String attr1, String attr2)
   {
      return 4+1+4 +4+(attr1.length()*2) +4+(attr2.length()*2);
   }
   
   public static int computeLRecLen(String data)
   {
      return 4+1+4 +4+(data.length()*2);
   }   
     
   public void updateFreeSpace(int reclen)
   {
      freeSpace -= reclen;
   }
   
  /**
   Delete a data record (in memmory) by set it's deleted bit to true.
   When this page is flushed to disk, but the deleted record is not 
   committed yet, the page will assume that this record is deleted.
   
   @param id of the deleting data record
   @return 0 if operation success will through exception is not success 
   */
   public int delete(int id)
   {
      IThNode node = getDataNode(id);
      node.setDeleted(true); 
    
      return 0; // success
   }
  
   /**
   This function update a data record with given id
    
   @param Id of the data record to be updated
   @param New value for the data record
   @param System where all the global variables hold.         
   @return Return Code of success (specified in properties file),
           Return Code of Error is updating value for the record 
           is not the same size as the old value. 
   */
   public String update(int id, String value, String lsn, ThSystem sys)
   {
      Properties prop = sys.getProperties();
    
      inFo.set(7); /* set dirty bit on */
      /* 1.locate data to be updated  */
      ThNode node = getDataNode(id); /* idx of the id's node */
      String data = (String) (node.getData()).get(0);  
      if (value.length() != data.length())
      {
         sys.out("Update data mismatched! "+
           prop.get("UPDATE_LENGTH_MISMATCH"));
        
         return (String) prop.get("UPDATE_LENGTH_MISMATCH");
      }  
      
      node.upDate(value); 
      
      return (String) prop.get("EC0");    
   }  
  
   /**
   Returns the data node for the given id.
    
   @param Id of the data node.
   @return:  Arraylist of values of the data node.
   */
   public ArrayList getData(int id)
   {
      ThNode node = getDataNode(id);
    
      return node.getData();
   }
 
   /**
   Compute to see if the given length of a record is fit into this page
    
   @param The length
   @return:  True if there the given data length is fit.
   @return:  False otherwise.
   */
   public boolean isFull(int length)
   {
      if (freeSpace - length > 0)
          return false;
      
      return true;
   }
     
   public int getSize()
   {
      return pSize;
   }
  
   /**
   This function returns the index for the data record of the given 
   record id
   @param id of the record
   @return:  Index of the node in this page.
   */
   public int getOffset(int id)
   {
      ThNode node;
    
      for (int i=0; i < nodeS.size(); i++)
      {
         node = (ThNode) nodeS.get(i);
         if (id == node.getId())
         return i;
      }
    
      return -1;
   }
  
   /**
   Returns the data node for the given id.
   @param Id of the data node.
   @return:  Arraylist of values of the data node.
   */
   public ThNode getDataNode(int id)
   {    
      int idx = getOffset(id);
      return (ThNode) nodeS.get(idx);
   }
         
   /**
   Remove a node in this page with a given id
   @param Id of the data node
   @return:  The removed record.
   */
   public IThNode remove(int id) 
   {
      int idx = getOffset(id);
      return (IThNode) nodeS.get(idx);
   }
  
   public int getNodeCount() 
   {
      return nodeS.size();
   }
   
   public IThNode getNodeByOffset(int idx) 
   {
      return (IThNode) nodeS.get(idx);
   }
  
   public void publicLogsToLM(ThSystem sys) 
   {
   }
  
   public void split(IThPage newpage, int id, Stack stack, ThSystem sys) 
   {
   }
  
   public void add(IThNode node) {
   }
   
   public void removeSubTree(ThGroup node) {
   }
   
   public void addSubTree(ThGroup node) {
   }
}