package cs297.xmldb.src.server;

import java.io.*;
import java.util.*;
import java.net.*;
import cs297.xmldb.src.cm.*;
import cs297.xmldb.src.sm.*;
import cs297.xmldb.src.bm.*;

/**
  COPYRIGHT (C) 2004 Thien An Nguyen. All Rights Reserved.
  This class is the execution engine of our toy DB.
 
  @Date: Spring & fall 2004
  @author: Thien An Nguyen, 
 */
public class ThAgent
{ 
   ThSystem syS;
 
 public ThAgent(ThSystem sub)
 {
    syS = sub;
 }
 
 /**
  If the return code in the plan is 2, return execution to the 
  caller right away as the DDL/DML input has user errors and it 
  need to be fixed before it can be re-submit.
  */
 public String executeDMLPlan(ThPlan plan, ArrayList resultSet)
 {
    Properties prop = syS.getProperties();
    String rc = plan.getReturnCode();
    String ec2 = (String)prop.get("EC2");
    boolean logYes = true;
   
    if (rc.equals(ec2))
       return ec2;  /* parsing error */
   
    syS.out("Executing DML Plan");  
    ArrayList stmts = plan.getParsedStmts();   
    for (int i = 0; i < stmts.size(); i++)  /* execute individual stmt */
    {
       ThStatement stmt = (ThStatement) stmts.get(i);   
       String command = stmt.getCommand();
       String txid = stmt.getTransactionId();
       syS.getTM().add(txid);

       if (command.equals("CRASH")) 
          rc = handleCrash();
       else if(command.equals("INSERT"))
          rc = handleInsert(stmt, txid, logYes);
       else if(command.equals("DELETE"))
          rc = handleDelete(stmt, txid, logYes, prop);
       else if(command.equals("UPDATE"))
          rc = handleUpdate(stmt, txid, logYes, prop);
       else if(command.equals("ROLLBACK"))
       {
          if (syS.isNatix())
             syS.getBM().publicLogs(syS);  
          syS.getRM().rollback(txid,syS);
       }
       else if(command.equals("COMMIT"))
          commitHandler(txid);
       else if(command.equals("LOAD"))
          rc = handleLoad(stmt, txid);
       else if(command.equals("CREATE XMLDOC"))
       {              
          String xmlname = (String)stmt.getParams().get(0);
          String dbname = (String)stmt.getParams().get(1);             
          rc = syS.createXmlDocEntry(txid,xmlname, dbname);
       }
       else if(command.equals("DROP XMLDOC"))  
          syS.dropXmlDoc(txid,(String)stmt.getParams().get(0)); 
       else if(command.equals("CREATE DATABASE"))
       {   
          String dbname = (String)stmt.getParams().get(0);
          rc = syS.createDB(txid,dbname);
       }
       else if(command.equals("DROP DATABASE"))
          rc = syS.dropDB(txid,(String)stmt.getParams().get(0));            
       else if(command.equals("PRINTTREE"))
          rc = handlePrintTree(stmt, txid);
       else if(command.equals("ARCHIVE"))
       {         
          String dbname = (String)stmt.getParams().get(0);
          archiveHandler(dbname);          
       }
       else /* unknown command, return statement error */
       {
          syS.out(command+" - "+ stmt.getErrorCode());
          rc = (String)prop.get(stmt.getErrorCode());
       }    
   }
   
   return rc;
 } 
 
 /**
  Handle for insert statement.
  It will look for the database object in the catalog
  Then locate the document in the database object
  and call delete from the document object.
  @params: stmt is the sql statement
           transaction id,
           log yes or no option
           properties object
           pointer to system object
  */ 
 public String handleInsert(ThStatement stmt, String txid, boolean lyes)
 {
    String xmlname = (String)stmt.getParams().get(0);
    /* look for current db in catalog  */
    String dbname = syS.getDB(xmlname);
    ThDB db = (ThDB)syS.cataLog.get(dbname);

    /* look for xmlname in current db */
    ThXMLDoc xmldoc = (ThXMLDoc)db.docNames.get(xmlname);    
    String path = (String)stmt.getParams().get(1);
    Object content = stmt.getParams().get(2);

    return xmldoc.insert(txid, path, content, syS, lyes);     
 }
 
 /**
  Handle for delete statement.
  It will look for the database object in the catalog
  Then locate the document in the database object
  and call delete from the document object.
  @params: stmt is the sql statement
           transaction id,
           log yes or no option
           properties object
           pointer to system object
  */ 
 public String handleDelete(ThStatement stmt, String txid, boolean lyes, 
                            Properties prop)
 {
    String xmlname = (String)stmt.getParams().get(0);
    /* look for current db in catalog */
    String dbname = syS.getDB(xmlname);
    ThDB db = (ThDB)syS.cataLog.get(dbname);
    /* look for xmlname in current db */
    ThXMLDoc xmldoc = (ThXMLDoc)db.docNames.get(xmlname);

    if (xmldoc == null)
    {
       syS.out(xmlname+" not found");
       return (String)prop.get("DOC_NOT_FOUND");
    }
    
    Object path = stmt.getParams().get(1);
    return xmldoc.delete(txid, path, syS, lyes); 
 }
 
  /**
  helper funtion as driver for update statement
  It will look for the database object in the catalog
  Then locate the document in the database object
  and call update from the document object.
  @params: stmt is the sql statement
           transaction id,
           log yes or no option
           properties object
           pointer to system object
  */ 
 public String handleUpdate(ThStatement stmt, String txid, 
                            boolean lyes, Properties prop)
 {
    String xmlname = (String)stmt.getParams().get(0);
    String newVal = (String)stmt.getParams().get(1);
    String path = (String)((ArrayList)stmt.getParams().get(2)).get(0);
    //look for current db in catalog       
    String dbname = syS.getDB(xmlname);
    ThDB db = (ThDB)syS.cataLog.get(dbname);
    //look for xmlname in current db 
    ThXMLDoc xmldoc = (ThXMLDoc)db.docNames.get(xmlname);
    if (xmldoc == null)
    {
       syS.out(xmlname+" not found");
       return (String)prop.get("DOC_NOT_FOUND");
    }
       
    return xmldoc.update(txid,xmlname, path, newVal, syS, lyes);    
 }
 
  /**
  Helper funtion as driver for load statement
  It will call load function handle in system object.
  @params: stmt is the sql statement
           transaction id,
  */  
 public String handleLoad(ThStatement stmt, String txid)
 {
    String xmlname = (String)stmt.getParams().get(1);
    String filename = (String)stmt.getParams().get(0);
    
    return syS.loadIntoXmlDoc(txid,filename,xmlname);      
 }
 
  /**
  helper funtion as driver for command commit 
  @params: pointer to system object
           transaction id
  */
 public int commitHandler(String txid)
 {
    if (syS.isNatix())
      syS.getBM().publicLogs(syS);  
    syS.getTM().commit(txid);
       
    return 0;
 }

  /**
  helper funtion as driver for command archive
  @params: pointer to system object
           the name of the database
  */ 
 public void archiveHandler(String dbname)
 {
    ThDB db = (ThDB)syS.cataLog.get(dbname); 
       
    if (syS.isNatix())
       syS.getBM().publicLogs(syS);
    
    //notice transaction table for the archive
    syS.getTM().archive();
    ThUtilities.imageCopy(db, syS);
    syS.out("Archive "+dbname);  
 }
 
 public String handleCrash()
 {
    syS.out("\nSimulating Crash!");
    syS.out("..........................");
    syS.out("Start Recovery from Crash......\n");

    return syS.recoverFromCrash();
 }
 
  /**
  helper funtion as driver for printTree command
  @params: pointer to system object
           the name of the database
  */  
 public String handlePrintTree(ThStatement stmt, String txid)
 {
    String xmlname = (String)stmt.getParams().get(0);
    String dbname = syS.getDB(xmlname);
    ThDB db = (ThDB)syS.cataLog.get(dbname);
    ThXMLDoc doc = (ThXMLDoc)db.docNames.get(xmlname);
    doc.printTree(txid, syS);
    
    return "000";
 }
}